//
//  MulleAutoreleasePool.m
//  MulleAutoreleasePool
//
//  Created by Nat! on Mon Dec 15 2003.
//  Copyright (c) 2003 Mulle kybernetiK. All rights reserved.
//

#define DEBUG        0
#define USE_POSER    0
#define USE_CATEGORY (! USE_POSER & 1)  // alternative


#import "MulleAutoreleasePool.h"

// WIN32 Port by Hervé Malaingre
#ifdef WIN32
#include <stdlib.h>
#include <windows.h>
#define pthread_mutex_t            CRITICAL_SECTION
#define pthread_mutex_init( X, Y)  InitializeCriticalSection(X)
#define pthread_mutex_lock( X)     EnterCriticalSection(X)
#define pthread_mutex_unlock( X)   LeaveCriticalSection(X)
#define pthread_self()             ((void *) GetCurrentThreadId())
#define pthread_t                  unsigned
#else
#include <pthread.h>
#endif
#include <objc/objc-runtime.h>


#define SELF  ((struct { @defs( NSAutoreleasePool) } *) self)

@implementation MulleAutoreleasePool


/*##
####
####     M   u   l   l   e   A   u   t   o   r   e   l   e   a   s   e   P   o   o   l   s   t   a   c  k
####
##*/
#define MULLE_NS_OBJECT_ARRAY_BYTES   4096    // one page
#define N_MULLE_NS_OBJECT_ARRAY       ((MULLE_NS_OBJECT_ARRAY_BYTES - sizeof( unsigned int) - sizeof( struct _MulleNSObjectArray *)) / sizeof( NSObject *))


typedef struct _MulleNSObjectArray
{
   NSObject                     *objects_[ N_MULLE_NS_OBJECT_ARRAY];
   unsigned int                 used_;
   struct _MulleNSObjectArray   *next_;
} MulleNSObjectArray;


typedef struct
{
   MulleAutoreleasePool   **pools_;
   unsigned int           size_;
   unsigned int           len_; 
   pthread_t              thread_;
} MulleAutoreleasePoolstack;


static void   addPoolToStack( MulleAutoreleasePoolstack *self, MulleAutoreleasePool *pool)
{
   if( self->len_ >= self->size_)
   {
      self->size_ += self->size_;
      if( self->size_ < 4)
         self->size_ = 4;
      self->pools_ = realloc( self->pools_, sizeof( MulleAutoreleasePool *) * self->size_);
   }
   self->pools_[ self->len_++] = pool;
}


static int   unwindAndReleasePoolsButDontReleaseLast(  MulleAutoreleasePoolstack *self, MulleAutoreleasePool *pool)
{
   int                    i;
   MulleAutoreleasePool   *p;
   
   for( i = self->len_ - 1; i >= 0; --i)
   {
      p = self->pools_[ i];
      if( p == pool)
      {
         self->len_ = i;
         return( i);
      }
      [p release];  // just call release here for safety. Could
                    // call dealloc but if the user REALLY wants
                    // to retain the pool, let him
   }
   return( 0);
}


/*##
####
####     M  u  l  l  e  A  u  t  o  r  e  l  e  a  s  e  P  o  o  l  s  t  a  c  k  S  t  o  r  a  g  e
####
##*/



typedef struct
{
   MulleAutoreleasePoolstack   **poolstacks_[ 63];
   unsigned int                sizes_[ 63];
   unsigned int                lens_[ 63];
   pthread_mutex_t             poolstacklock_;
   BOOL                        poolstacklock_initted_;
} MulleAutoreleasePoolstackStorage;


MulleAutoreleasePoolstackStorage   storage;


static inline unsigned int   hash_thread_modulo( pthread_t thread)
{
   return( (unsigned long) thread % 63);  // ok gimme a hash
}


static void   lock( MulleAutoreleasePoolstackStorage *self)
{
   if( ! self->poolstacklock_initted_)
   {
      pthread_mutex_init( &self->poolstacklock_, NULL);  // how can this fail ? should deal with it regardless
      self->poolstacklock_initted_ = YES;
   }
   pthread_mutex_lock( &self->poolstacklock_);
}


static inline void   unlock( MulleAutoreleasePoolstackStorage *self)
{
   pthread_mutex_unlock( &self->poolstacklock_);
}



/* not locking! */
static void   _addPoolstackForThreadIndex( MulleAutoreleasePoolstackStorage *self, MulleAutoreleasePoolstack *poolstack, unsigned int index)
{
   unsigned int   size;
   unsigned int   len;
   
   len  = self->lens_[ index];
   size = self->sizes_[ index];
   
   if( len >= size)
   {
      size += size;  /* double it up */
      if( ! size)
         size = 4;
      
      self->poolstacks_[ index] = realloc( self->poolstacks_[ index], size * sizeof( MulleAutoreleasePoolstack *));
      self->sizes_[ index]      = size;
   }
   
   self->poolstacks_[ index][ len] = poolstack;
#if DEBUG
   fprintf( stderr, "poolstack %p added at position %d in bucket %d\n", poolstack, len, index);
#endif   
   self->lens_[ index] = ++len;
}


static int   _indexOfPoolstackForThreadIndex( MulleAutoreleasePoolstackStorage *self, MulleAutoreleasePoolstack *poolstack, unsigned int index)
{
   MulleAutoreleasePoolstack   **stacks;
   unsigned int                i;
   unsigned int                len;
   
   stacks = self->poolstacks_[ index];
   len    = self->lens_[ index];
   
   for( i = 0; i < len; i++)
      if( stacks[ i] == poolstack)
         return( i);
   return( -1);
}


static int   _removePoolstackForThreadIndex( MulleAutoreleasePoolstackStorage *self, MulleAutoreleasePoolstack *poolstack, unsigned int index)
{
   MulleAutoreleasePoolstack   **stacks;
   unsigned int                len;
   int                        i;
   
   i = _indexOfPoolstackForThreadIndex( self, poolstack, index);
   if( i < 0)
      return( i);
   
   len    = self->lens_[ index] - 1;
   stacks = self->poolstacks_[ index];
   
   stacks[ i]          = 0; // remove
   self->lens_[ index] = len;
   
#if DEBUG
   fprintf( stderr, "poolstack %p removed from position %d of %d entries in bucket %d\n", poolstack, i, len + 1, index);
#endif   

   if( i < len)    // not tail removal 
   {   // need to memcpy argh!
      memcpy( &stacks[ i], 
              &stacks[ i + 1], 
              sizeof( MulleAutoreleasePoolstack **) * len - i);
   }
   return( 0);
}


static void   addPoolstackForThread( MulleAutoreleasePoolstackStorage *self, MulleAutoreleasePoolstack *poolstack, pthread_t thread)
{
   unsigned int   index;
   
   lock( self);
   
   index  = hash_thread_modulo( thread);  // ok gimme a hash
   _addPoolstackForThreadIndex( self, poolstack, index);
   
   unlock( self);
}


static void   removePoolstackForThread( MulleAutoreleasePoolstackStorage *self, MulleAutoreleasePoolstack *poolstack, pthread_t thread)
{
   unsigned int   index;
   
   lock( self);
   
   index  = hash_thread_modulo( thread);  // ok gimme a hash
   _removePoolstackForThreadIndex( self, poolstack, index);
   
   unlock( self);
}


static MulleAutoreleasePoolstack   *poolstackForThread( MulleAutoreleasePoolstackStorage *self, pthread_t thread)
{
   MulleAutoreleasePoolstack   **stacks;
   MulleAutoreleasePoolstack   *p;
   unsigned int                index;
   unsigned int                len;
   unsigned int                i;
   
   index  = hash_thread_modulo( thread);  // ok gimme a hash
   stacks = self->poolstacks_[ index];
   len    = self->lens_[ index];
   
   for( i = 0; i < len; i++)
      if( (p = *stacks++)->thread_ == thread)
         return( p);
   return( 0);
}


/*##
####
####     M   u   l   l   e   A   u   t   o   r   e   l   e   a   s   e   P   o   o   l  
####
##*/
static MulleAutoreleasePool   *poolForThread( pthread_t thread)
{
   MulleAutoreleasePoolstack   *p;
   
   if( p = poolstackForThread( &storage, thread))
      return( p->pools_[ p->len_ - 1]);
   return( nil);
}


static MulleAutoreleasePool   *poolForCurrentThread()
{
   return( poolForThread( pthread_self()));
}


static void    addPoolForThread( MulleAutoreleasePool *pool, pthread_t thread)
{
   MulleAutoreleasePoolstack   *p;
   
   if( ! (p = poolstackForThread( &storage, thread)))
   {
      p = malloc( sizeof( MulleAutoreleasePoolstack));
      memset( p, 0, sizeof( MulleAutoreleasePoolstack));
      p->thread_ = thread;
      addPoolstackForThread( &storage, p, thread);
   }
   
   addPoolToStack( p, pool);
}


static void    addPoolForCurrentThread( MulleAutoreleasePool *pool)
{
   addPoolForThread( pool, pthread_self());
}


/* this is probably needlessly slow.
 * As the search is done twice
 */
static void   unwindStackUntilPoolForThread( MulleAutoreleasePool *pool, pthread_t thread)
{
   MulleAutoreleasePoolstack   *p;
   
   if( ! (p = poolstackForThread( &storage, thread)))  
   {
      NSLog( @"*** Orphaned NSAutoreleasepool in pthread %p/NSThread %@. This indicates a retain/release problem *** ", 
             pthread_self(), 
             [NSThread currentThread]);
      return;
   }
   if( ! unwindAndReleasePoolsButDontReleaseLast( p, pool))
   {
      /* unwind all, then throw away poolstack */
      removePoolstackForThread( &storage, p, thread);
      free( p);
   }
}


static void   unwindStackUntilPoolForCurrentThread( MulleAutoreleasePool *pool)
{
   return( unwindStackUntilPoolForThread( pool, pthread_self()));
}



static inline Class   _isa( NSObject *obj)
{
   return( ((struct { @defs( NSObject) } *) obj)->isa);
}


//
//
// Experimenting with less:
// ading more lines slows down many many small pools
// decreasing the number increases chances that two isas clash
//
#define N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_    16  /* play with these */
#define N_MULLE_AUTORELEASE_CLASS_MASK_SHIFT_     0   


/* don't mess with these */
#define N_MULLE_AUTORELEASE_CLASS_PRESHIFT_       4   /* keep this constant */
#define N_MULLE_AUTORELEASE_CLASS_SHIFT_          (N_MULLE_AUTORELEASE_CLASS_PRESHIFT_ + N_MULLE_AUTORELEASE_CLASS_MASK_SHIFT_)

/* this does the right thing for us. It's not in an offical header
 * though. It looks through the cache first and refills the cache if
 * neccessary it only a little slower than objc_msgSend.
 * class_getInstanceMethod does the wrong thing in this case and is
 * too slow
 */
extern IMP   class_lookupMethod( Class cls, SEL sel);


static void   mulleAutoreleasePoolReleaseAll( MulleAutoreleasePool *self)
{
   MulleNSObjectArray   *array;

   IMP            lastRelease[ N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_];
   Class          lastIsa[ N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_];
   Class          thisIsa;
   NSObject       **rover;
   NSObject       *p;
   unsigned int   index;
   unsigned int   i;
   unsigned int   len;
   SEL            releaseSelector;
   IMP            (*p_class_lookupMethod)( Class isa, SEL sel);

   p_class_lookupMethod = class_lookupMethod;
   
   memset( lastIsa, 0, sizeof( lastIsa));
   releaseSelector = @selector( release);
   for( array = (MulleNSObjectArray *) SELF->_token; array; array = array->next_)
   {
      len   = array->used_;
      rover = array->objects_;
      
      for( i = 0; i < len; i++)
      {
         p       = *rover++;
         thisIsa = _isa( p);
       
         /* gcc probably optimizes this itself, but "kontrolle ist besser" */
#if ! (N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_ & (N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_ - 1))
         index = ((unsigned long) thisIsa >> N_MULLE_AUTORELEASE_CLASS_SHIFT_) & (N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_ - 1); 
#else         
         index = ((unsigned long) thisIsa >> N_MULLE_AUTORELEASE_CLASS_SHIFT_) % N_MULLE_AUTORELEASE_CLASS_CACHE_SIZE_; 
#endif         
         
         if( lastIsa[ index] != thisIsa)
         {
            if( ! (lastRelease[ index] = (*p_class_lookupMethod)( thisIsa, releaseSelector)))
               [NSException raise:NSGenericException
                           format:@"No release method on class %s", thisIsa->name];
            lastIsa[ index] = thisIsa;
         }
         (*lastRelease[ index])( p, releaseSelector);
      }
   }
}


/* NSAutoreleasePool seems to do some funky stuff in alloc (?) */

+ (id) alloc
{
   return( NSAllocateObject( self, 0, NULL));
}


+ (id) allocWithZone:(NSZone *) zone
{
   return( NSAllocateObject( self, 0, zone));
}


- (id) init
{
   MulleNSObjectArray  *array;
   
   // do not call NSAutoreleasePool
   addPoolForCurrentThread( self);  // memorize myself

   array = malloc( sizeof( MulleNSObjectArray));
   array->used_ = 0;
   array->next_ = 0;  
   
   SELF->_token = (void *) array;
   
   return( self); 
}


- (void) dealloc
{
   MulleNSObjectArray  *p, *q;
   
   unwindStackUntilPoolForCurrentThread( self); // does da popping
   
   mulleAutoreleasePoolReleaseAll( self);

   for( p = (MulleNSObjectArray *) SELF->_token; p; p = q)
   {
      q = p->next_;
      free( p);
   }

   NSDeallocateObject( self); // do not call NSAutoreleasePool
}


- (id) retain
{
   unsigned int  retainCount_;
   
   retainCount_  = (unsigned int) SELF->_reserved;
   SELF->_reserved = (void *) (retainCount_ + 1);
   return( self);
}


- (void) release
{
   unsigned int  retainCount_;
   
   retainCount_  = (unsigned int) SELF->_reserved;
   if( ! retainCount_)
   {
      [self dealloc];
      return;
   }
   SELF->_reserved = (void *) (retainCount_ - 1);
}


+ (NSString *) description
{
   return( [NSString stringWithFormat:@"<Class %@ %p>", NSStringFromClass( self), self]);  // hmm autoreleased... clever ?
}


- (NSString *) description
{
   return( [NSString stringWithFormat:@"<%@ %p>", NSStringFromClass( isa), self]);  // hmm autoreleased... clever ?
}


void   mulleAutoreleasePoolAdd( MulleAutoreleasePool *self, NSObject *p)
{
   MulleNSObjectArray   *array;
   
   array = (MulleNSObjectArray *) SELF->_token;
   if( array->used_ == N_MULLE_NS_OBJECT_ARRAY)
   {
      MulleNSObjectArray   *newArray;
      
      newArray = malloc( sizeof( MulleNSObjectArray));
      newArray->used_ = 0;
      newArray->next_ = array;
      
      SELF->_token = newArray;
      
      array = newArray;
   }
   array->objects_[ array->used_++] = p;
}   


void  mulleAutoreleasePoolLeakDetected()
{
   fprintf( stderr, 
            "*** No Autorelease pool in place for thread %p, just leaking\n*** Break on %s for debugging.\n",  
            pthread_self(),
            __PRETTY_FUNCTION__);
}
   

void   mulleAutoreleasePoolAddToCurrentPool( id anObject, SEL ignored)
{
   MulleAutoreleasePool   *pool;
   
   pool = poolForCurrentThread();   // slow as hell
   if( ! pool)
   {
      mulleAutoreleasePoolLeakDetected();
      return;
   }
   mulleAutoreleasePoolAdd( pool, anObject);
}


- (void) addObject:(id) anObject
{
   mulleAutoreleasePoolAdd( self, anObject);
}


+ (void) addObject:(id) anObject
{
   mulleAutoreleasePoolAddToCurrentPool( anObject, _cmd);  // wrong selector, doesn't matter
}


/* next two methodes courtesy of Hervé MALAINGRE
 * slightly reformatted to fit my coding style
 */
- (void) dumpPoolObjects
{
   MulleNSObjectArray   *array;
   unsigned int         i;
   
   array = (MulleNSObjectArray *) SELF->_token;

   fprintf( stderr, "Content of pool @ %p\n:", self);
   
   while( array) 
   {
      for( i = 0 ; i < array->used_ ; i++) 
         fprintf( stderr, 
                  "  - %s @ %p\n", 
                  object_getClassName( array->objects_[ i]), 
                  array->objects_[i]);
      array = array->next_;
   }
}

#if USE_POSER
+ (void) initializeMullePoolMechanism
{
   Method   originalMethod;
   
   originalMethod = class_getInstanceMethod( [NSObject class], 
                                             @selector(autorelease));
   [self poseAsClass:[NSAutoreleasePool class]];
   originalMethod->method_imp = (IMP) mulleAutoreleasePoolAddToCurrentPoolMethod;
}


+ (void) load
{
   [self initializeMullePoolMechanism];
}
#endif

@end

#if USE_CATEGORY

/* only if posing */
@interface NSObject ( MulleAutorelease)

- (id) autorelease;

@end


@implementation NSObject ( MulleAutorelease)

/* 
 * Would be nice to move this somewhere else, but we lose the inlining
 * potential. And it gets inlined!
 */
- (id) autorelease
{
   mulleAutoreleasePoolAddToCurrentPool( self, _cmd);
   return( self);
}

@end

#endif
